/***************************************************************************
 * Speextalk
 * (C) Nicolas George - 2005-07-24
 * This code is distributed under the terms of the GNU GPL.
 * sound_alsa.h: ALSA interface
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <alloca.h>
#include <alsa/asoundlib.h>

#include <sys/time.h>

#include "speextalk.h"

static snd_pcm_t *rec_pcm;
static snd_pcm_t *play_pcm;

/*
 * Dies printing the ALSA error message.
 */
static void
alsa_fatal(const char *msg, int e)
{
    const char *t;

    t = snd_strerror(e);
    if(t == NULL)
	t = "unknown error";
    fprintf(stderr, "%s: %s\n", msg, t);
    exit(1);
}

/*
 * Opens a PCM stream on the device name for recording or playing according
 * the the boolean rec. Dies on failure. Sample rate must be exact.
 */
static snd_pcm_t *
setup_pcm(const char *name, int rec, int sample_rate, int channels)
{
    snd_pcm_t *pcm;
    snd_pcm_hw_params_t *params;
    short endian = 0x1234;
    int e;

    pthread_mutex_lock(&global_mutex);
    if((e = snd_pcm_open(&pcm, name,
	rec ? SND_PCM_STREAM_CAPTURE : SND_PCM_STREAM_PLAYBACK,
	SND_PCM_NONBLOCK)))
	alsa_fatal(rec ? "snd_pcm_open(capture)" : "snd_pcm_open(playback)", e);
    pthread_mutex_unlock(&global_mutex);
    snd_pcm_hw_params_alloca(&params);
    snd_pcm_hw_params_any(pcm, params);
    snd_pcm_hw_params_set_access(pcm, params,
	SND_PCM_ACCESS_RW_INTERLEAVED);

    if(sizeof(endian) == 2 &&
	((unsigned char *)&endian)[0] == 0x12 &&
	((unsigned char *)&endian)[1] == 0x34) {
	snd_pcm_hw_params_set_format(pcm, params, SND_PCM_FORMAT_S16_BE);
    } else if(sizeof(endian) == 2 &&
	((unsigned char *)&endian)[0] == 0x34 &&
	((unsigned char *)&endian)[1] == 0x12) {
	snd_pcm_hw_params_set_format(pcm, params, SND_PCM_FORMAT_S16_LE);
    } else {
	fprintf(stderr, "Unsupported endianness\n");
	exit(1);
    }

    snd_pcm_hw_params_set_channels(pcm, params, channels);
    snd_pcm_hw_params_set_rate(pcm, params, sample_rate, 0);
    if((e = snd_pcm_hw_params(pcm, params)))
	alsa_fatal("snd_pcm_hw_params", e);
    return(pcm);
}

/*
 * Opens the recording PCM.
 */
void
prepare_recording(int sample_rate, int channels)
{
    int e;

    rec_pcm = setup_pcm(capture_device, 1, sample_rate, channels);
    snd_pcm_nonblock(rec_pcm, 0);
    if((e = snd_pcm_start(rec_pcm)))
	alsa_fatal("encode: snd_pcm_start", e);
}

/*
 * Opens the playing PCM.
 */
void
prepare_playing(int sample_rate, int channels)
{
    play_pcm = setup_pcm(playback_device, 0, sample_rate, channels);
    snd_pcm_nonblock(play_pcm, 0);
}

/*
 * Reads some samples from the recording PCM. Retries on recoverable errors.
 * Dies on fatal errors. Returns the number of samples read. Blocks.
 * Note: this API is not yet specified for multi-channels streams.
 */
int
read_samples(short *buf, int buf_size)
{
    int n;

    while(1) {
	if((n = snd_pcm_readi(rec_pcm, buf, buf_size)) < 0) {
	    if((n = snd_pcm_prepare(rec_pcm)) < 0)
		alsa_fatal("snd_pcm_readi: snd_pcm_prepare", n);
	    continue;
	}
	return(n);
    }
}

/*
 * Plays some samples on the playing PCM. Retries on recoverable errors.
 * Dies on fatal errors. Returns the number of samples written. Blocks.
 * Note: this API is not yet specified for multi-channels streams.
 */
int
play_samples(short *buf, int buf_size)
{
    int n;

    while(1) {
	if((n = snd_pcm_writei(play_pcm, buf, buf_size)) < 0) {
	    if((n = snd_pcm_prepare(play_pcm)) < 0)
		alsa_fatal("snd_pcm_writei: snd_pcm_prepare", n);
	    continue;
	}
	return(n);
    }
}
